#================================================================================================================================
# MIT LICENCE
# Copyright (c) 2019, Aziza Merzouki
#  
#  Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction, including without limitation 
#  the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, 
#  and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
#  
#  The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
#   INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
#   IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
#   WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE 
#   OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
#================================================================================================================================

library(clValid)
library(factoextra)
library(FactoMineR)
library(dendsort)

source("code/plotIncidenceMap.R")

#####################################################

# Perform hierarchical clustering based on a distance object

hierarchicalClustering = function(d,clusteringMethod){
  hc = hclust(d,method = clusteringMethod)
  return(hc)
}

#####################################################

# Function computeEuclidianDistance computes the Euclidian Distance between the Data Points
# Each Line is a Data Point, ex. a Vectorized Matrix

computeEuclidianDistance = function(M){
  n = dim(M)[1]
  euclidDist = array(0,dim=c(n,n))
  for (i in c(1:(n-1))){
    for (j in c((i+1):n)){
      diff = M[i,]-M[j,]
      dist = sqrt(sum(diff^2))
      euclidDist[i,j] = dist
      euclidDist[j,i] = dist
    }
  }
  rownames(euclidDist) = rownames(M)
  colnames(euclidDist) = rownames(M)
  return(euclidDist)
}

#####################################################

# Draw Map where Countries are coloured based on the Cluster they belong to.
# Inputs: 
# 1) Clusters (element name = Names of Country + element value = Cluster), 
# 2) Map Region, 
# 3) Map Title, 
# 4) Name of file where the map is saved 

drawMapOfClusters = function (clusters,region,title,fileName){
  pdf(fileName)
  # quartz()
  # Get map from rworldmap
  sPDF = getMap(resolution="high")  
  mapData = data.frame(country = names(clusters), cluster=clusters)
  #print(mapData)
  spdf = joinCountryData2Map(mapData,joinCode='NAME',nameJoinColumn="country")
  myPalette= "heat" #"rainbow"#
  mapCountryData(spdf,nameColumnToPlot="cluster",catMethod="categorical", mapRegion="africa",mapTitle=title,colourPalette=myPalette)
  # Add Country Labels on the Map
  labelCountries(spdf,nameCountryColum="country",cex=0.4,col="black")
  dev.off()
}

#####################################################

# Draw Box plots per cluster for a given attribute (given attribute File).

drawAttributeBoxplotsPerCluster = function(attributeFile,attributeName,countries,year,clusters,title,fileName){
  nClusters = max(clusters)
  print("drawAttributeBoxplotsPerCluster()")
  pdf(fileName)
  allCountryAttributes = getCountryAttribute(attributeFile, countries, year)
  allCountryAttributes = allCountryAttributes[paste("X",year,sep="")] # Ignore "Country" Columns
  print("BEFORE FIRST POINT PLOT")
  plot(1,1,col="white",xlab="Clusters",ylab=paste("Country",attributeName,year),main=title,xlim=c(1,nClusters),ylim=c(min(allCountryAttributes),max(allCountryAttributes)))
  print("BEFORE FIRST POINT PLOT")
  for(i in c(1:nClusters)){
    clusterCountryNames = names(clusters[clusters == i])
    clusterAttributes = getCountryAttribute(attributeFile, clusterCountryNames, year)
    print(paste(attributeName,"of Cluster",i))
    print(clusterAttributes[c("Country",paste("X",year,sep=""))])
    print(paste(attributeName,"summary of Cluster",i))
    print(summary(clusterAttributes))
  
    boxplot(clusterAttributes[paste("X",year,sep="")],add=TRUE,at=i)
  }
  dev.off()
}

#####################################################

# Draw Density Plots per cluster for all variables in matrix M.

draw_density_plots = function(M, clusters,mainDir,cutRefValue=""){
  extension = '.pdf'
  # extension = '.png'
  i = 1
  # for (f in rownames(M)){
  for (f in colnames(M)){
    fileName = paste(mainDir,'/density_var_',i,'_',cutRefValue,'Clusters',extension,sep='')
    pdf(fileName)
    # png(fileName)
    dat_f = c()
    dat_c = c()
    for (c in unique(clusters)){
      # Density
      dat_f = c(dat_f, M[clusters==c,f])
      dat_c = c(dat_c, rep(toString(c),length(M[clusters==c,f])))
    }
    
    dat = data.frame(feature=dat_f,cluster=dat_c)
    g = ggplot(dat, aes(x = feature, fill = cluster)) + 
      # geom_histogram(aes(y=..density..),alpha=0.5,position="identity")+
      geom_density(alpha = 0.5) +
      labs(x = f) +
      # scale_fill_manual(values=c("yellow", "orange", "red")) +
      theme(axis.text=element_text(size=14),
            axis.title=element_text(size=18),
            legend.text=element_text(size=14))
    plot(g)
    dev.off()
    i = i+1
  }
}

#####################################################
# Dimensionality Reduction of Data
# ex. the DAGs with N variables are represented by (N*N-N)/2 values => Vectorized DAGs.
# For N = 12, dim = 66. Use PCA to reduce it to 2-3 dimensions.
# reduceDimensions = function(allDAGs,clusters=NULL){
reduceDimensions = function(vectDAGs,names,clusters=NULL,filename,dir="code/Results"){
  
  # Transpose matrix to get "rows=countries" and "columns=variables"
  print("Data Matrix BEFORE PCA")
  print(vectDAGs)
  
  print(clusters)
  # PCA
  res.pca = prcomp(vectDAGs)
  
  # Print Rotation Matrix = Eigenvectors (the norm of each column = 1)
  print("res.pca = ")
  print(res.pca$rotation[,1:2])
  # Print Eigenvalues (Explained Variance = Eigenvalue/sum of eigenvalues)
  print(get_eigenvalue(res.pca))
  
  txtFilename = gsub('pdf','csv',filename)
  # write.table(res.pca,txtFilename)
  
  # Bar Chart of Explained Variance per PC
  explainedVar_filename = gsub('.pdf','_explainedVar.pdf',filename)
  pdf(explainedVar_filename)
  plot(fviz_eig(res.pca,addlabels = TRUE,linecolor = "#FC4E07",barcolor = "#00AFBB", barfill = "#00AFBB"))
  dev.off()
  
  print("Clusters = ")
  print(clusters)
  # Plot data point projections on PCA 2D-space
  pdf(filename)
  if (is.null(clusters)){
    plot(fviz_pca_ind(res.pca, repel=TRUE))
  } else{
    # plot(fviz_pca_ind(res.pca,col.ind=as.factor(clusters),legend.title = "Groups", addEllipses=FALSE, ellipse.type="confidence", repel=TRUE, palette =c("#FFDF00", "orange", "red")))
    plot(fviz_pca_ind(res.pca,col.ind=as.factor(clusters),legend.title = "Groups", addEllipses=TRUE, ellipse.type="confidence", repel=TRUE))
  }
  dev.off()
  
  # Checking PCs 
  # see ref: http://www.sthda.com/english/articles/31-principal-component-methods-in-r-practical-guide/112-pca-principal-component-analysis-essentials/
  var = get_pca_var(res.pca)
  
  print("eigen values")
  eig_val = get_eigenvalue(res.pca)
  print(eig_val)
  print("Check explained variance = eigenvalue/sum(eigenvalues)")
  print(eig_val[,"eigenvalue"]/sum(eig_val[,"eigenvalue"]))
  # Loadings = Rotation Matrix = Eigen Vectors
  print("Rotation matrix = Loadings")
  print(res.pca$rotation[,1:2])
  write.xlsx(res.pca$rotation[,1:2],paste(dir,"/loadings.xlsx",sep=""))
  
  print("Standard Deviations of Components = SQRT(eigenvalue(PC))")
  print(sqrt(eig_val[,"eigenvalue"]))
  print("var$coord = ")
  print(var$coord[,1:2])
  print("var$coord/eigenvector = std(PCs) ")
  print(var$coord[,1:2]/res.pca$rotation[,1:2])
  print("var$cor = ")
  print(var$cor[,1:2])
  print("var$cos2 = ")
  print(var$cos2[,1:2])
  print("var$contrib = ")
  print(var$contrib[,1:2])

  print("Eigen Value 2")
  print(eig_val[2,1])
  
  print("Contrib to (C1 * Eig1 + C2 * Eig2)/(Eig1+Eig2)")
  print((var$contrib[,1]*eig_val[1,"eigenvalue"] + var$contrib[,2]*eig_val[2,"eigenvalue"])/(eig_val[1,"eigenvalue"]+eig_val[2,"eigenvalue"]))
  print("Sum of contributions to PC1 = ")
  print(sum(var$contrib[,1]))

  # Plot initial variable contributions to PCs on PCA 2D-space
  contribution_percent = 2
  variablesContribution_filename = gsub('.pdf',paste0('_varContribution_',contribution_percent,'.pdf'),filename)
  
  print(paste("Number of variables with contribution to PC1 or PC2 > ", contribution_percent, "% = "))
  nbVar_contrib = sum(var$contrib[,1]>=contribution_percent | var$contrib[,2]>=contribution_percent)
  print(nbVar_contrib)

  # Visualize contribution of variables to PCs
  pdf(variablesContribution_filename)
  plot(fviz_pca_var(res.pca, geom = c("point","arrow", "text"), arrowsize= 0.2, labelsize = 3, alpha.var = "contrib", select.var = list(contrib = nbVar_contrib), repel=TRUE))
  dev.off()
  
  # Biplot - Top variable contribution to PC1 and PC2
  nbVar_biplot = 15
  biplot_filename = gsub('.pdf',paste0('_biplot_',nbVar_biplot,'.pdf'),filename)
  pdf(biplot_filename)
  plot(fviz_pca_biplot(res.pca, geom.ind = c("point","text"), geom.var = c("point","arrow", "text"), arrowsize= 0.3, label="all", labelsize = 3, alpha.var = "contrib", select.var = list(contrib = nbVar_biplot), repel=TRUE))
  dev.off()
  
  # Top variable contribution to PC1 and PC2
  contribution_top = 21 #26 #48
  variablesContribution_filename_top = gsub('.pdf',paste0('_topVarContrib_',contribution_top,'.pdf'),filename)
  pdf(variablesContribution_filename_top)
  plot(fviz_contrib(res.pca, choice = "var", axes = 1:2, top = contribution_top))
  dev.off()
  
  # Top variable contribution to PC1
  contribution_top = 21 #26 #48
  variablesContribution_filename_top = gsub('.pdf',paste0('_topVarContrib_1_',contribution_top,'.pdf'),filename)
  pdf(variablesContribution_filename_top)
  plot(fviz_contrib(res.pca, choice = "var", axes = 1, top = contribution_top))
  dev.off()
  # Top variable contribution to PC2
  contribution_top = 21 #26 #48
  variablesContribution_filename_top = gsub('.pdf',paste0('_topVarContrib_2_',contribution_top,'.pdf'),filename)
  pdf(variablesContribution_filename_top)
  plot(fviz_contrib(res.pca, choice = "var", axes = 2, top = contribution_top))
  dev.off()
  
  threeD_projection = predict(res.pca,vectDAGs)
  
  if (is.null(clusters)){
    Dim1 = threeD_projection[,1]
    Dim2 = threeD_projection[,2]
    Dim3 = threeD_projection[,3]
    scatterplot3d::scatterplot3d(Dim1,Dim2,Dim3,pch = 16,labels=rownames(threeD_projection))
  } else {

    # Cluster Colors
    col <- c("#CD5C5C","#999999", "#E69F00", "#56B4E9","#800080","#FF00FF","#000080","#0000FF","#008080","#00FFFF")
    colors <- col[as.numeric(clusters)]
    
    # Plot Projected Data (but keep z=0 for debugging)
    threeD_filename = gsub('.pdf','_3D_z0.pdf',filename)
    pdf(threeD_filename)
    scatterplot3d::scatterplot3d(x=threeD_projection[,1],y=threeD_projection[,2], z = matrix(0,dim(threeD_projection)[1],1),pch = 16,color=colors)
    # scatterplot3d::scatterplot3d(x=threeD_projection[,1],y=threeD_projection[,2], z = matrix(0,dim(threeD_projection)[1],1),pch = 16)
    
    legend("bottom", legend = levels(as.factor(clusters)),
           col = col,
           pch = c(16, 16, 16),
           inset = -0.05, xpd = TRUE, horiz = TRUE)
    dev.off()
    
    # Plot Projected Data (but keep z=0 for debugging)
    threeD_filename = gsub('.pdf','_3D.pdf',filename)
    pdf(threeD_filename)
    scatterplot3d::scatterplot3d(x=threeD_projection[,1],y=threeD_projection[,2], z = threeD_projection[,2],pch = 16)
    # scatterplot3d::scatterplot3d(x=threeD_projection[,1],y=threeD_projection[,2], z = threeD_projection[,2],pch = 16,color=colors)
    # legend("bottom", legend = levels(as.factor(clusters)),
    #        col = col, 
    #        pch = c(16, 16, 16),
    #        inset = -0.05, xpd = TRUE, horiz = TRUE)
    dev.off()
  }
  
  print("Done with DIM RED mehod!!!")
  return(threeD_projection)
}

#####################################################

# Get Feature Labels from the CSV file given as input.

getFeatureLabels = function(descriptionDir,jointGenders = FALSE){
  # Get feature labels
  print(descriptionDir)
  M=read.csv(descriptionDir)
  featureLabels = M[,"label_dhs"]
  featureLabels = gsub("\\s+\n\\s+"," ",featureLabels)
  
  if (jointGenders){
    otherGenderDir = ""
    # Male or Female Datasets
    if (grepl("Female",descriptionDir)==TRUE){
      featureLabels = paste(featureLabels,'women')
      otherGenderDir = gsub("Female","Male", descriptionDir)
      M=read.csv(otherGenderDir)
      otherGenderFeatureLabels = M[,"label_dhs"]
      otherGenderFeatureLabels = paste(otherGenderFeatureLabels,'men')
    } else {
      if (grepl("Male",descriptionDir)==TRUE){
        featureLabels = paste(featureLabels,'men')
        otherGenderDir = gsub("Male","Female", descriptionDir)
        M=read.csv(otherGenderDir)
        otherGenderFeatureLabels = M[,"label_dhs"]
        otherGenderFeatureLabels = paste(otherGenderFeatureLabels,'women')
      } else {
        print("ERROR: No Male or Female Datasets Directory.")
        return(-1)
      }
    }
    otherGenderFeatureLabels = gsub("\\s+\n\\s+"," ",otherGenderFeatureLabels)
    featureLabels = c(featureLabels,otherGenderFeatureLabels)
  }
  
  return(featureLabels)
}

