#================================================================================================================================
# MIT LICENCE
# Copyright (c) 2019, Aziza Merzouki
#  
#  Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files
# (the "Software"), to deal in the Software without restriction, including without limitation 
#  the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, 
#  and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
#  
#  The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, 
#   INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. 
#   IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
#   WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE 
#   OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
#================================================================================================================================

source("code/featuresClustering.R")

#####################################################

# Evaluate clustering using Silhouette Width and Dunn Index

evaluateClustering = function(d, clusters){
	# Silhouette width
	sil = silhouette(clusters,dist=d)
	mean_sil = mean(sil[,"sil_width"])
	# Dunn Index
	dunnIdx= dunn(d,clusters)
	
	return(list(mean_sil,dunnIdx))
}

#####################################################

# Perform Clustering using different numbers of clusters and Evaluate their results

runClusteringInvestigation = function(additionalDataFile, cutReference, cutRefValues, clusteringMethod="complete",incidenceYear=NULL,scale=FALSE){
  print("runClusteringInvestigation()")
  print(paste("additionalDataFile = ",additionalDataFile))
  print(paste("cutReference = ",cutReference))
  print(paste("cutRefValues = ",cutRefValues))
  print(paste("clusteringMethod = ",clusteringMethod))
  print(paste("incidenceYear = ",incidenceYear))
  
  # Cutting reference values 
  nRefValues = length(cutRefValues)
  allSil = array(dim = nRefValues)
  allDunn = array(dim = nRefValues)
  
  # Cluster with different Cutting Reference Values (height or number of clusters)
  for (i in c(1: nRefValues)){
		cutVal = cutRefValues[i]
		
		clustRes = runFeaturesClustering(additionalDataFile, cutReference, cutVal, clusteringMethod, incidenceYear, scale)
		
		clusters = clustRes[[1]]
		d = clustRes[[2]]	
		DS_id = clustRes[[3]]
		
		nClusters = max(clusters)
		nObs = length(clusters)
		if (nClusters > 1){
			evalRes = evaluateClustering(d, clusters)	
			allSil[i] = evalRes[[1]]
			allDunn[i] = evalRes[[2]]
		}	
	}
	
	max_sil = allSil[which.max(allSil)]
	optCutRefVal_sil = cutRefValues[which.max(allSil)]
	max_dunn = allDunn[which.max(allDunn)]
	optCutRefVal_dunn = cutRefValues[which.max(allDunn)]
	
	# Plot the Evolution of the Average Silhouette Width with respect to the Cutting Reference Value (Height or Number of Clusters)
	
	# Set directory where results must be saved.
	mainDir = "code/Results"
	dataType = "Features"
	if (scale){
	  fileName = paste(mainDir,'/',dataType,'_scale_sil_',DS_id,'_',cutReference,'_',clusteringMethod,'.pdf',sep='')
	} else {
	  fileName = paste(mainDir,'/',dataType,'_sil_',DS_id,'_',cutReference,'_',clusteringMethod,'.pdf',sep='')
	}
	pdf(fileName)
	plot(cutRefValues,allSil,xlab=cutReference,ylab="Average Silhouette Width",main=DS_id)
	legend("topright",legend = paste("max_sil = ", format(max_sil,digits=2), " - ", cutReference, " = ", optCutRefVal_sil))
	dev.off()
	
	return(list(max_sil, optCutRefVal_sil,max_dunn, optCutRefVal_dunn))
}
